/*
 Quasar-mini interface board Linux driver.

 Copyright (C) 2007 Parabel Ltd.
 Author: Mikhail Khelik <mixel.net@gmail.com>

 Copyright (C) 2008-2011 Parabel Ltd.
 Author: Pavel Selivanov <pavel[AT]parabel.ru>

 * Parabel Ltd. reserve the right to distribute project or it's parts
 under the alternative conditions.

 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software Foundation;
 either version 2 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 See the GNU General Public License for more details.

 You should have received a copy of the GNU General Public License along with this program;
 if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/ioctl.h>
#include <linux/poll.h>
#include <linux/cdev.h>
#include <linux/device.h>
#if defined(HAVE_UNLOCKED_IOCTL) && defined(CONFIG_BKL)                                                                                                         
# include <linux/smp_lock.h> /* lock_kernel */
#endif
#include <linux/version.h>

#include "kal.h"
#include "version.h"
#include "pmc_e1_conf.h"
#include "pmc_e1.h"
#include "e12zap.h"
#include "pmc_e1_if.h" //include after <linux/ioctl.h>

/* Defines */
#define DRV_NAME "quasarm"
#define WIN_SIZE 0x10000 /* PCI window size */

/* Global variables */
static dev_t dev_file_glob; //descriptop for devfs and sysfs
static unsigned char device_number = 0;
static unsigned int cacheline = 0x20;
static unsigned int cacheline_bridge = 0x20;
static unsigned int latency = 0;
static int irqflags = IRQF_SHARED | IRQF_DISABLED;
LIST_HEAD(dev_list_head);

/* Global initialized structs */
static struct pci_device_id ptmc_e1_tbl[] = {
	{ 0x1172, 0x2230, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
	{ 0, }
};
MODULE_DEVICE_TABLE(pci, ptmc_e1_tbl);

/* Function prototypes */
/* */

#ifdef PMC_E1_UDEV
static struct class *pmc_e1_class = NULL;
#else
# define ptmc_sysfs_create(a)
# define ptmc_sysfs_delete(a)
#endif //PMC_E1_UDEV

/*========== Common functions ==========*/
static void init_ptmc_regs(const struct pmc_e1_controller * tmp_dev)
{
	tmp_dev->bar0[0x02] = 0xe0000000 | MODE1_MASL_HP | MODE1_CLR_ZERO_HP | MODE1_CT_TI_EN_HP | MODE1_CONS_DEL_HP ; //Configure board
	udelay(1000);
	tmp_dev->bar0[0x02] = 0x60000000 | MODE1_MASL_HP | MODE1_CLR_ZERO_HP | MODE1_CT_TI_EN_HP | MODE1_DMARST_HP | MODE1_CONS_DEL_HP; //Wake up framers, hm magic reset
	udelay(1000);
	//tmp_dev->bar0[0x10] = 2048; //Configure RTC for 1000 interrupts/s
	tmp_dev->bar0[0x01] = 0x80000000;	
	tmp_dev->bar0[0x06] = 0x00000003; // BS/LS: allow to sync to all E1 channels;
}

static void init_switch_map(const struct pmc_e1_controller * tmp_dev)
{
    int i;
//Mapping of the timeslots for the Quasar Mini
//////////From E1 to PC///////////////////////////////////////////////
    //LINE2(DMA WRITE BUFFER(quasar->pc)) <- LINE0(E1_0 input)
    for(i=0;i<32;i++) {
	tmp_dev->bar0[SWITCH_TABLE_OFFSET + 0x1200+i] = INPUT_TS_SIGNATURE(0, i);
    }
    //LINE3(DMA WRITE BUFFER(quasar->pc)) <- LINE1(E1_1 input)
    for(i=0;i<32;i++) {
	tmp_dev->bar0[SWITCH_TABLE_OFFSET + 0x1300+i] = INPUT_TS_SIGNATURE(1, i);
    }
/////////From PC to E1////////////////////////////////////////////////
    //LINE0(E1_0 output) <- LINE2(DMA READ BUFFER(quasar->pc))
    for(i=0;i<32;i++) {
	tmp_dev->bar0[SWITCH_TABLE_OFFSET + 0x1000+i] = INPUT_TS_SIGNATURE(2, i);
    }
    //LINE1(E1_1 output) <- LINE3(DMA READ BUFFER(quasar->pc))
    for(i=0;i<32;i++) {
	tmp_dev->bar0[SWITCH_TABLE_OFFSET + 0x1100+i] = INPUT_TS_SIGNATURE(3, i);
    }
//////////////////////////////////////////////////////////////////////

}

int acc_lxt_reg(struct pmc_e1_controller * tmp_dev ,unsigned char chan_num, unsigned char addr, unsigned char mode, unsigned char *data) {

    if (chan_num > 1) return -1;
    if (mode > 1) return -1;
    if (!tmp_dev) return -1;
    if (addr > 7) return -1;


    tmp_dev->bar0[LIU_CTRL] = 0;
    tmp_dev->bar0[LIU_CTRL] = (chan_num << 10) | LIU_CS; //set cs for definite channel;
    addr = (addr | 0x10 ) << 1;
    tmp_dev->bar0[LIU_CTRL] = (chan_num << 10) | LIU_CS | LIU_TR | addr | mode;  //pass address and operation
    while (tmp_dev->bar0[LIU_CTRL] & LIU_TR) {};
    if (LIU_WRITE == mode) {
        tmp_dev->bar0[LIU_CTRL] = ((chan_num) << 10) | LIU_CS | LIU_TR | *data;
        while (tmp_dev->bar0[LIU_CTRL] & LIU_TR) {};
    } else {
        tmp_dev->bar0[LIU_CTRL] = ((chan_num) << 10) | LIU_CS | LIU_TR;
        while (tmp_dev->bar0[LIU_CTRL] & LIU_TR) {};
        *data = 0xff & tmp_dev->bar0[LIU_CTRL];
    }
    tmp_dev->bar0[LIU_CTRL] = 0;
    return 0;
}

static void init_framer_regs(struct pmc_e1_controller * tmp_dev){
    unsigned char data, i; //, j;
    //LIU IO test;
    /*
    for(i=0; i < 8; i++ ) {
	for (j=0; j < 255; j++ ) {
		data = j;
		module_dprintk(KERN_DEBUG, "i=%d, Write to LIU: 0x%02x ", i, data);
		acc_lxt_reg(tmp_dev, 0, i, LIU_WRITE, &data);
		data =0;
		acc_lxt_reg(tmp_dev, 0, i, LIU_READ, &data);
		module_dprintk(KERN_DEBUG, "Read from LIU: 0x%02x \n", data);
	}

    } */   


    for (i=0; i <= tmp_dev->e1_amount; i++ ){
        data = 1;
        acc_lxt_reg(tmp_dev, i, LR_CR1, LIU_WRITE, &data);
        data = 0;
        acc_lxt_reg(tmp_dev, i, LR_CR1, LIU_READ, &data);
        if (1 == data) {
	    module_printk(KERN_INFO, "%s: LIU found on port %d\n", tmp_dev->name, i);
            tmp_dev->bar1[(0x40*i)+FRC1_ADDR_OFFSET] = FRC_FR_RX_EN | FRC_FR_TX_EN | FRC_CRC4_TX_EN |
                FRC_MFR_TX_EN | FRC_EBIT_TX1 | FRC_EBIT_TX0;
            data = LR_CR2_RESET;
            acc_lxt_reg(tmp_dev, i, LR_CR2, LIU_WRITE, &data);
            mdelay(35);
	    data = LR_CR1_JASEL0 | LR_CR1_UNIEB | LR_CR1_ENCENB | LR_CR1_EC1 | LR_CR1_EC4; //HDB3, long haul, JA-tx
          //  printk("Data for LR_CR1 = 0x%02x\n", data);
	    acc_lxt_reg(tmp_dev, i, LR_CR1, LIU_WRITE, &data);
            data = 0;
            acc_lxt_reg(tmp_dev, i, LR_CR1, LIU_READ, &data);
	    module_dprintk(KERN_DEBUG, "LIU CR1 register = 0x%02x\n", data);
	    data = LR_CR4_CODEV | LR_CR4_ZEROV;
            acc_lxt_reg(tmp_dev, i, LR_CR4, LIU_WRITE, &data);

        } else {
	    module_printk(KERN_INFO, "%s: Using short haul mode on port %d\n", tmp_dev->name, i);
            tmp_dev->bar1[(0x40*i)+FRC1_ADDR_OFFSET] = FRC_FR_RX_EN | FRC_FR_TX_EN | FRC_CRC4_TX_EN |
                FRC_MFR_TX_EN | FRC_EBIT_TX1 | FRC_EBIT_TX0 | FRC_HDB3_RX_EN | FRC_HDB3_TX_EN |
                FRC_SHORT_HAUL_EN | FRC_TXEN;
        }
	module_dprintk(KERN_DEBUG, "Framer config register: 0x%x\n", tmp_dev->bar1[(0x40*i)]);
	
	tmp_dev->bar1[(0x40*i) + CAS_BIT_BUFFER_OFFSET] = 0x0b;
        //Set transmit 5 on abcd bits for all available channels
	
	for (data=1; data<16; data++){
		tmp_dev->bar1[(0x40*i) + CAS_BIT_BUFFER_OFFSET + data] = 0x55;
	}
    }

}

/*========== IOCTL (ecfg) ==========*/
/*
 static void check_bit(..)
 to_hw - structure to pass configuration
 from_hw - structure to return actual settings from hardware back
 smask - position of property bit in configuration structure
 hmask - position of same bit in hardware register
*/

static void check_bit_m(struct pmc_e1_controller * tmp_dev,
			unsigned char chan_num, unsigned char smask,
			unsigned int hmask, unsigned char clear_hwbit_if_swbit_one,
			unsigned char to_hw, unsigned char * from_hw)
{
    unsigned int tmp;
    tmp = tmp_dev->bar1[0x40*chan_num];
    tmp &= ~hmask; //clear bit;
    if(to_hw & smask){
        if (clear_hwbit_if_swbit_one){

        } else { //set bit
            tmp |= hmask;
        }

    } else {
        if (clear_hwbit_if_swbit_one){
            tmp |= hmask;
        } else {

        }
    }
    tmp_dev->bar1[0x40*chan_num] = tmp;
    //read back
    tmp = 0;
    tmp = tmp_dev->bar1[0x40*chan_num];
    if(tmp & hmask)
    {
        if (clear_hwbit_if_swbit_one){

        } else { //set bit
            *from_hw |= smask;
        }

    } else {
        if (clear_hwbit_if_swbit_one){
            *from_hw |= smask;
        } else {
            
        }
    }
    return;
}

static void align_to_swposition(struct pmc_e1_controller * tmp_dev, unsigned char chan_num, unsigned char smask,
                                unsigned int hmask, unsigned char clear_hwbit_if_swbit_one, unsigned char * from_hw) {
    unsigned int tmp=0;
    tmp = tmp_dev->bar1[0x40*chan_num];
    if(tmp & hmask)
    {
        if (clear_hwbit_if_swbit_one){

        } else { //set bit
            *from_hw |= smask;
        }

    } else {
        if (clear_hwbit_if_swbit_one){
            *from_hw |= smask;
        } else {
            
        }
    }
    return;
}

/*
 Returns:
 ENOTTY - in case of invalid ioctl;
 EINVAL - in case of passing invalid parameters;
 EFAULT - in case of error in the data transmitin beetween user and kernel space;
 ENOSYS - Function not implemented
 0 - in case of success;
*/

static int pmc_e1_ioctl(struct inode * inode_p, struct file * file_p, unsigned int cmd, unsigned long arg)
{
	int err = 0;
	unsigned int tmp=0, j;
        unsigned char ucTmp=0;
	struct pmc_config_chan store_chanl_conf, read_back_chan_conf;
	struct pmc_err_ins err_ins;
	struct pmc_data_pass pass_data;
	struct pmc_status_chan stat_exchange;
	struct pmc_config_board board_config;
        struct pmc_cas_xmit_data cas_data;
	struct pmc_version version_exchange;
	struct pmc_e1_controller * tmp_dev = (struct pmc_e1_controller *) file_p->private_data;

	memset((void *) &board_config, 0, sizeof(struct pmc_config_board));
	memset((void *) &read_back_chan_conf, 0, sizeof(struct pmc_config_chan));
	memset((void *) &store_chanl_conf, 0, sizeof(struct pmc_config_chan));
        memset((void *) &err_ins, 0, sizeof(struct pmc_err_ins));
	memset((void *) &stat_exchange, 0, sizeof(struct pmc_status_chan));
        memset((void *) &cas_data, 0, sizeof(struct pmc_cas_xmit_data));
	if (_IOC_TYPE(cmd) != PTMC_E1_8_MAGIC)
		return -ENOTTY;
	if (_IOC_NR(cmd) > PTMC_E1_8_MCN)
		return -ENOTTY;
	if (_IOC_DIR(cmd) & _IOC_READ)
		err = !access_ok(VERIFY_WRITE, (void *) arg, _IOC_SIZE(cmd));
	else if (_IOC_DIR(cmd) & _IOC_WRITE)
		err = !access_ok(VERIFY_READ, (void *) arg, _IOC_SIZE(cmd));
	if(err)
		return -EFAULT;

	switch(cmd) {
	case PTMC_E1_CHAN_CONFIGURE:
		module_dprintk(KERN_DEBUG, "PTMC_E1_CHAN_CONFIGURE.\n");
		if(copy_from_user((void *) &store_chanl_conf, (void *) arg, sizeof(store_chanl_conf)))
		{
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		if ((store_chanl_conf.chan_num & (~DIRECT_IN)) > tmp_dev->e1_amount) return -EINVAL;

		module_dprintk(KERN_DEBUG, "Before config, chan:%d data: %x\n", store_chanl_conf.chan_num, tmp_dev->bar1[0x40 * store_chanl_conf.chan_num]);

		if (!(store_chanl_conf.chan_num & DIRECT_IN)) //writing cas data
		{
			store_chanl_conf.chan_num&= (~DIRECT_IN);

			//cfg0

			//tmp_dev, chan_num,                           , smask   ,     hmask,         Nbit, to_hw                ,  from_hw
			check_bit_m(tmp_dev, store_chanl_conf.chan_num, CFG0_CAS,     FRC_MFR_TX_EN, 1   , store_chanl_conf.cfg0,  &read_back_chan_conf.cfg0 );
			check_bit_m(tmp_dev, store_chanl_conf.chan_num, CFG0_RCV_TR,  FRC_FR_RX_EN,  1   , store_chanl_conf.cfg0,  &read_back_chan_conf.cfg0 );
			check_bit_m(tmp_dev, store_chanl_conf.chan_num, CFG0_XMIT_TR, FRC_FR_TX_EN , 1   , store_chanl_conf.cfg0,  &read_back_chan_conf.cfg0 );

			tmp = tmp_dev->bar1[0x40 * store_chanl_conf.chan_num];

			if (tmp & FRC_SHORT_HAUL_EN) { //short haul
				check_bit_m(tmp_dev, store_chanl_conf.chan_num, CFG0_RCV_HDB3, FRC_HDB3_RX_EN, 0, store_chanl_conf.cfg0,  &read_back_chan_conf.cfg0);
				check_bit_m(tmp_dev, store_chanl_conf.chan_num, CFG0_XMIT_HDB3, FRC_HDB3_TX_EN, 0, store_chanl_conf.cfg0,  &read_back_chan_conf.cfg0);
			} else {
				module_dprintk(KERN_DEBUG, "Configure HDB3 on long haul port\n"); //TODO,  ! on framer only both tx + rx
				if((store_chanl_conf.cfg0 & CFG0_RCV_HDB3) || (store_chanl_conf.cfg0 & CFG0_XMIT_HDB3)) { //if HDB3 on RX or TX enable HDB3 on LIU
					ucTmp = LR_CR1_JASEL0 | LR_CR1_UNIEB | LR_CR1_ENCENB | LR_CR1_EC1 | LR_CR1_EC4; //HDB3, long haul, JA-tx;
					module_dprintk(KERN_DEBUG, "To write ucTmp =0x%02x ch_num=%d\n", ucTmp, store_chanl_conf.chan_num);
					acc_lxt_reg(tmp_dev, store_chanl_conf.chan_num, LR_CR1, LIU_WRITE, &ucTmp);
					ucTmp = 0;
					acc_lxt_reg(tmp_dev, store_chanl_conf.chan_num, LR_CR1, LIU_READ, &ucTmp);
					module_dprintk(KERN_DEBUG, "Read back ucTmp =0x%02x ch_num=%d\n", ucTmp, store_chanl_conf.chan_num);
					if (ucTmp & LR_CR1_ENCENB) {
						read_back_chan_conf.cfg0 |= CFG0_RCV_HDB3;
						read_back_chan_conf.cfg0 |= CFG0_XMIT_HDB3;
					} else {
						module_printk(KERN_ERR, "LIU configuration error\n");
					}
				} else {
					module_dprintk(KERN_DEBUG, "Configure AMI on long haul port\n"); //TODO,  ! on framer only both tx + rx
					ucTmp = LR_CR1_JASEL0 | LR_CR1_UNIEB | LR_CR1_EC1 | LR_CR1_EC4;
					module_dprintk(KERN_DEBUG, "To write ucTmp =0x%02x ch_num=%d\n", ucTmp, store_chanl_conf.chan_num);
					acc_lxt_reg(tmp_dev, store_chanl_conf.chan_num, LR_CR1, LIU_WRITE, &ucTmp);
					ucTmp = 0;
					acc_lxt_reg(tmp_dev, store_chanl_conf.chan_num, LR_CR1, LIU_READ, &ucTmp);
					module_dprintk(KERN_DEBUG, "Read back ucTmp =0x%02x ch_num=%d\n", ucTmp, store_chanl_conf.chan_num);
					if (ucTmp & LR_CR1_ENCENB)
					{
						read_back_chan_conf.cfg0 |= CFG0_RCV_HDB3;
						read_back_chan_conf.cfg0 |= CFG0_XMIT_HDB3;
						module_printk(KERN_ERR, "LIU configuration error\n");
					}

				}
			}

			//configure board part
			if (store_chanl_conf.cfg0 & CFG0_MASTER)
			{
				tmp = tmp_dev->bar0[0x06]; //get bs/ls
				ucTmp = (1<<store_chanl_conf.chan_num);
				tmp|= (unsigned int) ucTmp;
				tmp_dev->bar0[0x06] = tmp; //restore bs/ls
				read_back_chan_conf.cfg0|= CFG0_MASTER;
			}
			else                                          //allow to sync to this chan
			{
				tmp = tmp_dev->bar0[0x06]; //get bs/ls
				ucTmp = ~(1<<store_chanl_conf.chan_num);
				tmp&= (unsigned int) ucTmp;
				tmp_dev->bar0[0x06] = tmp; //restore bs/ls
			}



			//CFG1
			check_bit_m(tmp_dev, store_chanl_conf.chan_num, CFG1_NAT0, FRC_SA_BITS0, 0, store_chanl_conf.cfg1,  &read_back_chan_conf.cfg1);
			check_bit_m(tmp_dev, store_chanl_conf.chan_num, CFG1_NAT1, FRC_SA_BITS1, 0, store_chanl_conf.cfg1,  &read_back_chan_conf.cfg1);
			check_bit_m(tmp_dev, store_chanl_conf.chan_num, CFG1_NAT2, FRC_SA_BITS2, 0, store_chanl_conf.cfg1,  &read_back_chan_conf.cfg1);
			check_bit_m(tmp_dev, store_chanl_conf.chan_num, CFG1_NAT3, FRC_SA_BITS3, 0, store_chanl_conf.cfg1,  &read_back_chan_conf.cfg1);
			check_bit_m(tmp_dev, store_chanl_conf.chan_num, CFG1_NAT4, FRC_SA_BITS4, 0, store_chanl_conf.cfg1,  &read_back_chan_conf.cfg1);
			check_bit_m(tmp_dev, store_chanl_conf.chan_num, CFG1_INAT0, FRC_EBIT_TX0, 0, store_chanl_conf.cfg1,  &read_back_chan_conf.cfg1);
			check_bit_m(tmp_dev, store_chanl_conf.chan_num, CFG1_INAT1, FRC_EBIT_TX1, 0, store_chanl_conf.cfg1,  &read_back_chan_conf.cfg1);


			//CFG2
			check_bit_m(tmp_dev, store_chanl_conf.chan_num, CFG2_RAI_MANUAL, FRC_ARAI, 0, store_chanl_conf.cfg2,  &read_back_chan_conf.cfg2);
			check_bit_m(tmp_dev, store_chanl_conf.chan_num, CFG2_RAI_ON, FRC_RAI_TX, 0, store_chanl_conf.cfg2,  &read_back_chan_conf.cfg2);
			//CFG2_RCAS_AUTO_ON //not implemented

			//CFG2_RCAS_ON
			if (store_chanl_conf.cfg2 & CFG2_RCAS_ON) { //enable MF alarm (Y -bit -> 1)
				tmp_dev->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] |= CAS_Y_BIT;
			} else {
				tmp_dev->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] &= ~CAS_Y_BIT;
			}
			//read back
			if (tmp_dev->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] & CAS_Y_BIT) {
				read_back_chan_conf.cfg2 |= CFG2_RCAS_ON;
			}

			//CFG2_X1
			if (store_chanl_conf.cfg2 & CFG2_X1) {
				tmp_dev->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] |= CAS_X1_BIT;
			} else {
				tmp_dev->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] &= ~CAS_X1_BIT;
			}
			//read back
			if (tmp_dev->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] & CAS_X1_BIT) {
				read_back_chan_conf.cfg2 |= CFG2_X1;
			}
			//CFG2_X2
			if (store_chanl_conf.cfg2 & CFG2_X2) {
				tmp_dev->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] |= CAS_X2_BIT;
			} else {
				tmp_dev->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] &= ~CAS_X2_BIT;
			}
			//read back
			if (tmp_dev->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] & CAS_X2_BIT) {
				read_back_chan_conf.cfg2 |= CFG2_X2;
			}
			//CFG2_X3
			if (store_chanl_conf.cfg2 & CFG2_X3) {
				tmp_dev->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] |= CAS_X3_BIT;
			} else {
				tmp_dev->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] &= ~CAS_X3_BIT;
			}
			//read back
			if (tmp_dev->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] & CAS_X3_BIT) {
				read_back_chan_conf.cfg2 |= CFG2_X3;
			}

			//CFG3

			//CFG3_CRC4_TR ignore
			check_bit_m(tmp_dev, store_chanl_conf.chan_num, CFG3_CRC4_DIS, FRC_CRC4_TX_EN, 1, store_chanl_conf.cfg3,  &read_back_chan_conf.cfg3);
			//CFG3_CRC4_AUTO ignore

			//CFG4
			check_bit_m(tmp_dev, store_chanl_conf.chan_num, CFG4_AIS,   FRC_TXAO, 1, store_chanl_conf.cfg4,  &read_back_chan_conf.cfg4);
			check_bit_m(tmp_dev, store_chanl_conf.chan_num, CFG4_RLOOP, FRC_RLBK, 0, store_chanl_conf.cfg4,  &read_back_chan_conf.cfg4);
			//check_bit_m(tmp_dev, store_chanl_conf.chan_num, CFG4_PLOOP, FRC_, 0, store_chanl_conf.cfg4,  &read_back_chan_conf.cfg4);

			tmp = tmp_dev->bar1[0x40 * store_chanl_conf.chan_num];

			if (tmp & FRC_SHORT_HAUL_EN) { //short haul
				check_bit_m(tmp_dev, store_chanl_conf.chan_num, CFG4_LLOOP, FRC_MLBK, 0, store_chanl_conf.cfg4,  &read_back_chan_conf.cfg4);
				check_bit_m(tmp_dev, store_chanl_conf.chan_num, CFG4_TX_DIS, FRC_TXEN, 1, store_chanl_conf.cfg4,  &read_back_chan_conf.cfg4);
			} else {
				if(store_chanl_conf.cfg4 & CFG4_LLOOP) {
					module_dprintk(KERN_DEBUG, "Enable local loopback on long haul port\n");
					ucTmp =	0;
					acc_lxt_reg(tmp_dev, store_chanl_conf.chan_num, LR_CR2, LIU_READ, &ucTmp);
					ucTmp |= LR_CR2_EALOOP;
					acc_lxt_reg(tmp_dev, store_chanl_conf.chan_num, LR_CR2, LIU_WRITE, &ucTmp);
					ucTmp = 0;
					acc_lxt_reg(tmp_dev, store_chanl_conf.chan_num, LR_CR2, LIU_READ, &ucTmp);
					if (ucTmp & LR_CR2_EALOOP) {
						read_back_chan_conf.cfg4 |= CFG4_LLOOP;
					} else {
						module_printk(KERN_ERR, "LIU configuration error\n");
					}
				} else {
					module_dprintk(KERN_DEBUG, "Disable local loopback on long haul port\n");
					ucTmp =	0;
					acc_lxt_reg(tmp_dev, store_chanl_conf.chan_num, LR_CR2, LIU_READ, &ucTmp);
					ucTmp &= ~LR_CR2_EALOOP;
					acc_lxt_reg(tmp_dev, store_chanl_conf.chan_num, LR_CR2, LIU_WRITE, &ucTmp);
					ucTmp = 0;
					acc_lxt_reg(tmp_dev, store_chanl_conf.chan_num, LR_CR2, LIU_READ, &ucTmp);
					if (ucTmp & LR_CR2_EALOOP) {
						module_printk(KERN_ERR, "LIU configuration error\n");
						read_back_chan_conf.cfg4 |= CFG4_LLOOP;
					}
				}
				//	check_bit_m(tmp_dev, store_chanl_conf.chan_num, CFG4_TX_DIS, FRC_TXEN, 1, store_chanl_conf.cfg4,  &read_back_chan_conf.cfg4);
			}



		}
		else //only read data;
		{
			store_chanl_conf.chan_num&= (~DIRECT_IN);


			//cfg0
			//tmp_dev, chan_num,                        , smask   ,     hmask,       , Nbit,  from_hw
			align_to_swposition(tmp_dev, store_chanl_conf.chan_num, CFG0_CAS,     FRC_MFR_TX_EN, 1   , &read_back_chan_conf.cfg0 );
			align_to_swposition(tmp_dev, store_chanl_conf.chan_num, CFG0_RCV_TR,  FRC_FR_RX_EN,  1   , &read_back_chan_conf.cfg0 );
			align_to_swposition(tmp_dev, store_chanl_conf.chan_num, CFG0_XMIT_TR, FRC_FR_TX_EN , 1   , &read_back_chan_conf.cfg0 );
			tmp = tmp_dev->bar1[0x40 * store_chanl_conf.chan_num];

			if (tmp & FRC_SHORT_HAUL_EN) { //short haul
				align_to_swposition(tmp_dev, store_chanl_conf.chan_num, CFG0_RCV_HDB3, FRC_HDB3_RX_EN, 0,  &read_back_chan_conf.cfg0);
				align_to_swposition(tmp_dev, store_chanl_conf.chan_num, CFG0_XMIT_HDB3, FRC_HDB3_TX_EN, 0, &read_back_chan_conf.cfg0);
			} else {
				ucTmp = 0;
				acc_lxt_reg(tmp_dev, store_chanl_conf.chan_num, LR_CR1, LIU_READ, &ucTmp);
				if (ucTmp & LR_CR1_ENCENB) {
					read_back_chan_conf.cfg0 |= CFG0_RCV_HDB3;
					read_back_chan_conf.cfg0 |= CFG0_XMIT_HDB3;
				}
			}
			tmp = tmp_dev->bar0[0x06]; //get bs/ls
			tmp = tmp >> store_chanl_conf.chan_num;
			if (tmp & 1) {
				read_back_chan_conf.cfg0 |= CFG0_MASTER;
			}

			//cfg1
			align_to_swposition(tmp_dev, store_chanl_conf.chan_num, CFG1_NAT0, FRC_SA_BITS0, 0,        &read_back_chan_conf.cfg1);
			align_to_swposition(tmp_dev, store_chanl_conf.chan_num, CFG1_NAT1, FRC_SA_BITS1, 0,        &read_back_chan_conf.cfg1);
			align_to_swposition(tmp_dev, store_chanl_conf.chan_num, CFG1_NAT2, FRC_SA_BITS2, 0,        &read_back_chan_conf.cfg1);
			align_to_swposition(tmp_dev, store_chanl_conf.chan_num, CFG1_NAT3, FRC_SA_BITS3, 0,        &read_back_chan_conf.cfg1);
			align_to_swposition(tmp_dev, store_chanl_conf.chan_num, CFG1_NAT4, FRC_SA_BITS4, 0,        &read_back_chan_conf.cfg1);
			align_to_swposition(tmp_dev, store_chanl_conf.chan_num, CFG1_INAT0, FRC_EBIT_TX0, 0,       &read_back_chan_conf.cfg1);
			align_to_swposition(tmp_dev, store_chanl_conf.chan_num, CFG1_INAT1, FRC_EBIT_TX1, 0,       &read_back_chan_conf.cfg1);

			//cfg2
			align_to_swposition(tmp_dev, store_chanl_conf.chan_num, CFG2_RAI_MANUAL, FRC_ARAI, 0,     &read_back_chan_conf.cfg2);
			align_to_swposition(tmp_dev, store_chanl_conf.chan_num, CFG2_RAI_ON, FRC_RAI_TX, 0,       &read_back_chan_conf.cfg2);

			//CFG2_RCAS_ENA
			if (tmp_dev->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] & CAS_Y_BIT) {
				read_back_chan_conf.cfg2 |= CFG2_RCAS_ON;
			}
			if (tmp_dev->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] & CAS_X3_BIT) {
				read_back_chan_conf.cfg2 |= CFG2_X3;
			}
			if (tmp_dev->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] & CAS_X2_BIT) {
				read_back_chan_conf.cfg2 |= CFG2_X2;
			}
			if (tmp_dev->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] & CAS_X1_BIT) {
				read_back_chan_conf.cfg2 |= CFG2_X1;
			}

			//CFG3

			align_to_swposition(tmp_dev, store_chanl_conf.chan_num, CFG3_CRC4_DIS, FRC_CRC4_TX_EN, 1,   &read_back_chan_conf.cfg3);


			//cfg4
			align_to_swposition(tmp_dev, store_chanl_conf.chan_num, CFG4_AIS   , FRC_TXAO, 	      1,   &read_back_chan_conf.cfg4);
			align_to_swposition(tmp_dev, store_chanl_conf.chan_num, CFG4_RLOOP , FRC_RLBK,        0,   &read_back_chan_conf.cfg4);
			//    	check_bit_m(tmp_dev, store_chanl_conf.chan_num, CFG4_PLOOP, FRC_, 0, store_chanl_conf.cfg4,  &read_back_chan_conf.cfg4);

			tmp = tmp_dev->bar1[0x40 * store_chanl_conf.chan_num];

			if (tmp & FRC_SHORT_HAUL_EN) { //short haul
				align_to_swposition(tmp_dev, store_chanl_conf.chan_num, CFG4_LLOOP, FRC_MLBK, 0,  &read_back_chan_conf.cfg4);
				align_to_swposition(tmp_dev, store_chanl_conf.chan_num, CFG4_TX_DIS, FRC_TXEN, 1, &read_back_chan_conf.cfg4);
			} else {
				ucTmp = 0;
				acc_lxt_reg(tmp_dev, store_chanl_conf.chan_num, LR_CR2, LIU_READ, &ucTmp);
				if (ucTmp & LR_CR2_EALOOP) read_back_chan_conf.cfg4 |= CFG4_LLOOP;
				//check_bit_m(tmp_dev, store_chanl_conf.chan_num, CFG4_TX_DIS, FRC_TXAO, 0, store_chanl_conf.cfg4,  &read_back_chan_conf.cfg4); //no equvivalent in LIU Mode
			}

		}
		module_dprintk(KERN_DEBUG, "After config, chan: %d data: %x\n", store_chanl_conf.chan_num, tmp_dev->bar1[0x40 * store_chanl_conf.chan_num]);
		if(copy_to_user((void *) arg, (void *) &read_back_chan_conf, sizeof(read_back_chan_conf)))
		{
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		break;
	case PTMC_E1_ERR_EMU:
		module_printk(KERN_WARNING, "PTMC_E1_ERR_EMU.\n");
		module_printk(KERN_WARNING, "Unimplemented.\n");
		return -ENOSYS;
		//AIS

		break;
	case PTMC_E1_GET_STATS:
		module_dprintk(KERN_DEBUG, "PTMC_E1_GET_STATS.\n");
		if(copy_from_user((void *) &stat_exchange, (void *) arg, sizeof(stat_exchange)))
		{
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		if (stat_exchange.chan_num > tmp_dev->e1_amount) return -EINVAL;

		//stat 0
		tmp = tmp_dev->bar1[(0x40 * stat_exchange.chan_num) + FRS_ADDR_OFFSET];
		stat_exchange.stat0 = 0;

		if (tmp & FRS_LOS) stat_exchange.stat0 |= STAT0_LOS;
		if (tmp & FRS_RED) stat_exchange.stat0 |= STAT0_LOF;
		if (!(tmp & FRS_MFRAME_SYNC)) stat_exchange.stat0 |= STAT0_CAS_LOM;
		if (!(tmp & FRS_CRC_SYNC)) stat_exchange.stat0 |= STAT0_CRC4_SYNC;
		if (!((tmp & FRS_EBIT_RX1) && (tmp & FRS_EBIT_RX0))) stat_exchange.stat0 |= STAT0_E_BIT;
		if (tmp & FRS_CRC4_ERR) stat_exchange.stat0 |= STAT0_CRC4;
		if (tmp & FRS_AIS) stat_exchange.stat0 |= STAT0_AIS;
		if (tmp & FRS_RAI_RX) stat_exchange.stat0 |= STAT0_RAIS;

		//stat1
		stat_exchange.stat1 = 0;
		if (tmp & FRS_SA_BITS0) stat_exchange.stat1 |= STAT1_NAT0;
		if (tmp & FRS_SA_BITS1) stat_exchange.stat1 |= STAT1_NAT1;
		if (tmp & FRS_SA_BITS2) stat_exchange.stat1 |= STAT1_NAT2;
		if (tmp & FRS_SA_BITS3) stat_exchange.stat1 |= STAT1_NAT3;
		if (tmp & FRS_SA_BITS4) stat_exchange.stat1 |= STAT1_NAT4;
		if (tmp & FRS_EBIT_RX0) stat_exchange.stat1 |= STAT1_INAT0;
		if (tmp & FRS_EBIT_RX1) stat_exchange.stat1 |= STAT1_INAT1;
		if (tmp & FRS_SLIP_ERR) stat_exchange.stat1 |= STAT1_SLIP;

		//stat2
		stat_exchange.stat2 = 0;
		if (tmp & FRS_MFRAI) stat_exchange.stat2 |= STAT2_Y;
		tmp = tmp_dev->bar1[(0x40 * stat_exchange.chan_num) + CAS_BIT_BUFFER_OFFSET];
		if (tmp & CAS_RX_X1_BIT) stat_exchange.stat2 |= STAT2_X1;
		if (tmp & CAS_RX_X2_BIT) stat_exchange.stat2 |= STAT2_X2;
		if (tmp & CAS_RX_X3_BIT) stat_exchange.stat2 |= STAT2_X3;

		//hdb3
		tmp = tmp_dev->bar1[(0x40 * stat_exchange.chan_num) + FRS_HDB3ERR_OFFSET];
		stat_exchange.hdb3 = (unsigned short) tmp & 0xFFFF;

		//crc4
		tmp = tmp_dev->bar1[(0x40 * stat_exchange.chan_num) + FRS_CRC4ERR_OFFSET];
		stat_exchange.crc4 = (unsigned short) tmp & 0xFFFF;

		//fas
		tmp = tmp_dev->bar1[(0x40 * stat_exchange.chan_num) + FRS_FRAMEERR_OFFSET];
		stat_exchange.fas = (unsigned short) tmp & 0xFFFF;

		//mgmt

		if (stat_exchange.mgmt & 0x01) tmp_dev->bar1[(0x40 * stat_exchange.chan_num) + FRS_HDB3ERR_OFFSET] = 0;
		if (stat_exchange.mgmt & 0x02) tmp_dev->bar1[(0x40 * stat_exchange.chan_num) + FRS_FRAMEERR_OFFSET] = 0;
		if (stat_exchange.mgmt & 0x04) tmp_dev->bar1[(0x40 * stat_exchange.chan_num) + FRS_CRC4ERR_OFFSET] = 0;

		//ovr
		stat_exchange.ovr = 0;

		//get cas signaling

		for (j=0; j<16; j=j+2) {
			tmp = (tmp_dev->bar1[(0x40 * stat_exchange.chan_num) + CAS_BIT_BUFFER_OFFSET + j] >> 8) & 0xf0;
			stat_exchange.cas_data[j/2] = (unsigned char) tmp;
			//	printk("j=%d, data=0x%x, i=%d, tmp = 0x%x\n", j, tmp_dev->bar1[(0x40 * stat_exchange.chan_num) + CAS_BIT_BUFFER_OFFSET + j] , j/2, tmp);
		}

		for (j=1; j<16; j=j+2) {
			tmp = (tmp_dev->bar1[(0x40 * stat_exchange.chan_num) + CAS_BIT_BUFFER_OFFSET + j] >> 12) & 0xf;
			stat_exchange.cas_data[j/2] |= (unsigned char) tmp;
			//	printk("j=%d, data=0x%x, i=%d, tmp = 0x%x\n", j, tmp_dev->bar1[(0x40 * stat_exchange.chan_num) + CAS_BIT_BUFFER_OFFSET + j] , j/2, tmp);
		}

		for (j=0; j<16; j=j+2) {
			tmp = (tmp_dev->bar1[(0x40 * stat_exchange.chan_num) + CAS_BIT_BUFFER_OFFSET + j] >> 4) & 0xf0;
			stat_exchange.cas_data[8 + j/2] = (unsigned char) tmp;
			//	printk("j=%d, data=0x%x, i=%d, tmp = 0x%x\n", j, tmp_dev->bar1[(0x40 * stat_exchange.chan_num) + CAS_BIT_BUFFER_OFFSET + j] , j/2, tmp);
		}

		for (j=1; j<16; j=j+2) {
			tmp = (tmp_dev->bar1[(0x40 * stat_exchange.chan_num) + CAS_BIT_BUFFER_OFFSET + j] >> 8) & 0xf;
			stat_exchange.cas_data[8 + j/2] |= (unsigned char) tmp;
			//	printk("j=%d, data=0x%x, i=%d, tmp = 0x%x\n", j, tmp_dev->bar1[(0x40 * stat_exchange.chan_num) + CAS_BIT_BUFFER_OFFSET + j] , j/2, tmp);
		}

		if(copy_to_user((void *) arg, (void *) &stat_exchange, sizeof(stat_exchange)))
		{
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		break;
	case PTMC_E1_RESET:
		module_printk(KERN_WARNING, "PTMC_E1_RESET.\n");
		module_printk(KERN_WARNING, "Unimplemented/n");
		return -ENOSYS;
		break;
	case PTMC_E1_BOARD_CONFIGURE:
		module_printk(KERN_WARNING, "PTMC_E1_BOARD_CONFIGURE.\n");
		module_printk(KERN_WARNING, "Unimplemented/n");
		return -ENOSYS;
		break;
	case PTMC_E1_GET_PCI_SLOT:
		module_dprintk(KERN_DEBUG, "PTMC_E1_GET_PCI_SLOT.\n");
		tmp = PCI_SLOT( (unsigned int) tmp_dev->pci_device->devfn);
		if (copy_to_user((void *) arg, (void *) &tmp, _IOC_SIZE(cmd)))
		{
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		break;
	case PTMC_E1_GET_PCI_BUSNUM:
		module_dprintk(KERN_DEBUG, "PTMC_E1_GET_PCI_BUSNUM.\n");
		tmp = tmp_dev->pci_device->bus->number;
		if (copy_to_user((void *) arg, (void *) &tmp, _IOC_SIZE(cmd)))
		{
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		break;
	case PTMC_E1_READ_BAR0:
		module_dprintk(KERN_DEBUG, "PTMC_E1_READ_BAR0.\n");
		if (copy_from_user((void *) &pass_data, (void *) arg, sizeof(pass_data)))
		{
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		if (pass_data.raddress > 0xFFFF) return -EINVAL;
		pass_data.rdata = tmp_dev->bar0[pass_data.raddress];
		if (copy_to_user((void *) arg, (void *) &pass_data, sizeof(pass_data)) )
		{
			return -EFAULT;
		}
		break;
	case PTMC_E1_WRITE_BAR0:
		module_dprintk(KERN_DEBUG, "PTMC_E1_WRITE_BAR0.\n");
		if (copy_from_user((void *) &pass_data, (void *) arg, sizeof(pass_data)))
		{
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		if (pass_data.raddress > 0xFFFF) return -EINVAL;
		tmp_dev->bar0[pass_data.raddress] = pass_data.rdata;
		break;
	case PTMC_E1_READ_BAR1:
		module_dprintk(KERN_DEBUG, "PTMC_E1_READ_BAR1.\n");
		if (copy_from_user((void *) &pass_data, (void *) arg, sizeof(pass_data)))
		{
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		if (pass_data.raddress > 0xFFFF) return -EINVAL;
		pass_data.rdata = tmp_dev->bar1[pass_data.raddress];
		if (copy_to_user((void *) arg, (void *) &pass_data, sizeof(pass_data)) )
		{
			return -EFAULT;
		}
		break;
	case PTMC_E1_WRITE_BAR1:
		module_dprintk(KERN_DEBUG, "PTMC_E1_WRITE_BAR1.\n");
		if (copy_from_user((void *) &pass_data, (void *) arg, sizeof(pass_data)))
		{
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		if (pass_data.raddress > 0xFFFF) return -EINVAL;
		tmp_dev->bar1[pass_data.raddress] = pass_data.rdata;
		break;
	case PTMC_E1_CAS_XMIT_PASS:
		module_dprintk(KERN_DEBUG, "PTMC_E1_CAS_XMIT_PASS\n");
		if(copy_from_user((void *) &cas_data, (void *) arg, _IOC_SIZE(cmd)))
		{
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}

		if ((cas_data.chan_num & (~DIRECT_IN)) > tmp_dev->e1_amount) return -EINVAL;

		if (copy_to_user((void *) arg, (void *) &cas_data, sizeof(cas_data)) )
		{
			return -EFAULT;
		}

		for (j=2; j<16; j=j+2) { //j==2 -> abcd2 -> cdata[1]
			tmp = tmp_dev->bar1[(0x40 * cas_data.chan_num) + CAS_BIT_BUFFER_OFFSET + j];
			tmp &= 0xff0f;
			tmp |= cas_data.cdata[j/2] & 0xf0;
			tmp_dev->bar1[(0x40 * cas_data.chan_num) + CAS_BIT_BUFFER_OFFSET + j] = tmp;
		}

		for (j=1; j<16; j=j+2) {
			tmp = tmp_dev->bar1[(0x40 * cas_data.chan_num) + CAS_BIT_BUFFER_OFFSET + j];
			tmp &= 0xff0f;
			tmp |= ((cas_data.cdata[j/2] & 0x0f) << 4);
			tmp_dev->bar1[(0x40 * cas_data.chan_num) + CAS_BIT_BUFFER_OFFSET + j] = tmp;
		}
		for (j=2; j<16; j=j+2) {
			tmp = tmp_dev->bar1[(0x40 * cas_data.chan_num) + CAS_BIT_BUFFER_OFFSET + j];
			tmp &= 0xfff0;
			tmp |= ((cas_data.cdata[8 + j/2] & 0xf0) >> 4);
			tmp_dev->bar1[(0x40 * cas_data.chan_num) + CAS_BIT_BUFFER_OFFSET + j] = tmp;

		}
		for (j=1; j<16; j=j+2) {
			tmp = tmp_dev->bar1[(0x40 * cas_data.chan_num) + CAS_BIT_BUFFER_OFFSET + j];
			tmp &= 0xfff0;
			tmp |= (cas_data.cdata[8 + j/2] & 0x0f);
			tmp_dev->bar1[(0x40 * cas_data.chan_num) + CAS_BIT_BUFFER_OFFSET + j] = tmp;
		}

		break;
	case PTMC_E1_GET_VERSION:
		module_dprintk(KERN_DEBUG, "PTMC_E1_VERSION.\n");
		strncpy (version_exchange.proto_version, ecfg_proto_version, sizeof (version_exchange.proto_version));
		version_exchange.hw_fw_rev = tmp_dev->bar0[0x0D];
		strncpy (version_exchange.driver_version, QUASARM_VERSION, sizeof (version_exchange.driver_version));
		if(copy_to_user((void *) arg, (void *) &version_exchange, sizeof(version_exchange)))
		{
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		break;



		/*
		 case PTMC_E1_PASS_ABCD_ZAP_FAKE:
		 //printk("PTMC_E1_PASS_ABCD_ZAP_FAKE\n");
		 if(copy_from_user((void *) &tmp, (void *) arg, _IOC_SIZE(cmd)))
		 {
		 module_printk(KERN_ERR, "Memory access fault!\n");
		 return -EFAULT;
		 }
		 tmp_dev->test_sig = tmp;
		 //printk("Passing:0x%x\n", tmp);
		 break;
		 */
	default:
		return -ENOTTY;
		break;
	}
	return 0;
}

#ifdef HAVE_UNLOCKED_IOCTL
#  if !defined(CONFIG_BKL)
#    define lock_kernel();
#    define unlock_kernel();
#  endif //CONFIG_BKL
static long pmc_e1_ioctl_unlocked(struct file *file, unsigned int cmd, unsigned long data)
{
	long res;

	lock_kernel();
	res = pmc_e1_ioctl(NULL, file, cmd, data);
	unlock_kernel();
        return res;
}
#endif //HAVE_UNLOCKED_IOCTL

static int pmc_e1_open(struct inode * inode_p, struct file * file_p)
{
	file_p->private_data = container_of(inode_p->i_cdev, struct pmc_e1_controller, dev_file);
	return 0;
}

static int pmc_e1_release(struct inode * inode_p, struct file * file_p)
{
	return 0;
}

static struct file_operations pmc_e1_fops  = {
    owner:		THIS_MODULE,
    open:		pmc_e1_open,
#ifdef HAVE_UNLOCKED_IOCTL
	unlocked_ioctl:	pmc_e1_ioctl_unlocked,
#else
	ioctl:		pmc_e1_ioctl,
#endif
    release:            pmc_e1_release,
};

/*========== SYSFS code ==========*/
#ifdef PMC_E1_UDEV
struct ptmc_attribute {
    struct device_attribute attr;
    unsigned long p[3];
};
#define toptattr(a) ((struct ptmc_attribute *)(a))
#define todeattr(a) ((struct device_attribute *)(a))

#define PTMC_ATTR_RO(name) __ATTR(name, S_IRUGO, _pmc_show, NULL)
ssize_t _pmc_show(struct device *cls, struct device_attribute *attr,
		      char *buff)
{
    struct pmc_e1_controller * tmp_dev;
    unsigned long *p = toptattr(attr)->p;

    tmp_dev = (struct pmc_e1_controller *) dev_get_drvdata(cls);
    return sprintf(buff,"%d\n", (tmp_dev->bar0[p[0]] >> p[1]) && p[2]);
}

ssize_t pmc_show(struct device *cls, struct device_attribute *attr, char *buff)
{
    int count=0,i;
    unsigned int tmp;
    struct pmc_e1_controller * tmp_dev;
    tmp_dev = (struct pmc_e1_controller *) dev_get_drvdata(cls);

    count+=sprintf(buff,"ChN:");
    for(i=0; i <= tmp_dev->e1_amount; i++)
	count+=sprintf(buff+count," %d",i);

    count+=sprintf(buff+count,"\nLOS:");
    for(i=0; i<=tmp_dev->e1_amount; i++)
    {
	tmp = tmp_dev->bar1[(0x40 * i) + FRS_ADDR_OFFSET];
	count+=sprintf(buff+count," %d", tmp & FRS_LOS ? 1 : 0);
    }

    spin_lock(&tmp_dev->dev_lock);
    count+=sprintf(buff+count,"\nLOF:");
    for(i=0; i<= tmp_dev->e1_amount; i++)
    {
	tmp = tmp_dev->bar1[(0x40 * i) + FRS_ADDR_OFFSET];
	count+=sprintf(buff+count," %d", tmp & FRS_RED ? 1 : 0);
    }
    count+=sprintf(buff+count,"\nLOM:");
    for(i=0; i <= tmp_dev->e1_amount;i++)
    {
	tmp = tmp_dev->bar1[(0x40 * i) + FRS_ADDR_OFFSET];
	count+=sprintf(buff+count," %d", !(tmp & FRS_MFRAME_SYNC) ? 1 : 0);
    }
    spin_unlock(&tmp_dev->dev_lock);

    count+=sprintf(buff+count,"\n");
    return count;
}

static struct ptmc_attribute device_attrs[] = {
    {PTMC_ATTR_RO(pmc_load), 	{0x21,16,0xFF}},
    {PTMC_ATTR_RO(pmc_load_max),{0x21,0,0xFF}},
    {PTMC_ATTR_RO(pci_load), 	{0x20,16,0xFF}},
    {PTMC_ATTR_RO(pci_load_cur),{0x20,0,0xFF}},
    {PTMC_ATTR_RO(dma_overflow),{0x09,24,0xFF}},
    {__ATTR(info, S_IRUGO, pmc_show, NULL), {0,}},
};

static int ptmc_sysfs_create (struct pmc_e1_controller * tmp_dev) {
    int i, error = 0;
    struct device *ptmc_device = tmp_dev->pmc_class_dev;

    tmp_dev->pmc_class_dev = ptmc_device =
	QUASAR_DEVICE_CREATE (pmc_e1_class, QUASAR_PARENT(tmp_dev),
		       MKDEV(MAJOR(dev_file_glob), tmp_dev->dev_number),
		       tmp_dev->name);
    if (IS_ERR(tmp_dev->pmc_class_dev)) {
        module_printk (KERN_WARNING, "Failed to device_create.\n");
	return 1;
    }
    dev_set_drvdata(tmp_dev->pmc_class_dev, (void *) tmp_dev);

    for (i = 0; i < ARRAY_SIZE(device_attrs); i++) {
	error = device_create_file(ptmc_device, &(device_attrs[i].attr));
	if (error)
	    break;
    }

    if (error) {
	while (--i >= 0)
	    device_remove_file(ptmc_device, todeattr(&device_attrs[i]));
    }

    return 0;
}

static void ptmc_sysfs_delete (struct pmc_e1_controller * tmp_dev) {
    int i;
    struct device *ptmc_device = tmp_dev->pmc_class_dev;

    if(IS_ERR(tmp_dev->pmc_class_dev)) return;

    for (i = 0; i < ARRAY_SIZE(device_attrs); i++)
	device_remove_file(ptmc_device, todeattr(&device_attrs[i]));

    device_destroy(pmc_e1_class, MKDEV(MAJOR(dev_file_glob), tmp_dev->dev_number));
}
#endif //PMC_E1_UDEV

/*========== Init code & IRQ handler ==========*/
QUASAR_IRQ_HANDLER(ptmc_e1_interrupt_hr)
{
	unsigned int flags=0;
	struct pmc_e1_controller * tmp_dev = (struct pmc_e1_controller *) pr_data;
	if(tmp_dev == NULL ) return IRQ_NONE;
	flags = tmp_dev->bar0[0];
        if (flags & 0x20000000) //check DMA interrupt;
        //if (flags & 0x40000000) //check RTC interrupt;
	{
		tmp_dev->bar0[0x0]=0x0;
	       	e12zap_delivery(tmp_dev);
		return IRQ_HANDLED;
	}
	return IRQ_NONE;
}

static int set_cacheline (struct pci_dev *pdev, u8 val) {
	u8 cache_tmp;

	/* Write the correct value. */
	pci_write_config_byte(pdev, PCI_CACHE_LINE_SIZE, val);
	/* Read it back. */
	pci_read_config_byte(pdev, PCI_CACHE_LINE_SIZE, &cache_tmp);
	if (val != cache_tmp) {
	    module_printk(KERN_ERR, "Cache line size of %d DWORDs is not "
			  "supported (%x:%x)\n", val,
			  pdev->vendor, pdev->device);
	    return -EINVAL;
	}

        return 0;
}

/*
 Some BIOS or Linux, set PCIe bridge "cache line size" to illegal value (>20).
 Setting "cache line size" for our FPGA & PCIe bridge (PEX 8111, PEX 8112).
 Is PCIe bridge the only device on the PCI bus (can I use first device) ?
*/
int tune_pci (struct pci_dev *pdev) {
	struct pci_bus *bus;
        struct pci_dev *bridge;
	struct list_head *next;
	u8 lat_timer, tmp_u8;

	bus = pdev->bus->parent;
        next = bus->devices.next;
	bridge = pci_dev_b (next);

 	module_printk (KERN_INFO, "Setting Cache Line Size to %d\n", cacheline);
	if (set_cacheline (pdev, cacheline))
            return -EINVAL;

	pci_read_config_byte(pdev, PCI_LATENCY_TIMER, &lat_timer);
	if (lat_timer < 40 || latency) {
		lat_timer = latency ? latency : 40;
		module_printk (KERN_INFO, "Setting latency timer to %d\n",
			       lat_timer);
		pci_write_config_byte(pdev, PCI_LATENCY_TIMER, lat_timer);
		pci_read_config_byte(pdev, PCI_LATENCY_TIMER, &tmp_u8);
		if (tmp_u8 != lat_timer) {
                        printk ("%d %d\n", tmp_u8, lat_timer);
			module_printk(KERN_ERR, "Latency timer of %d is not "
				      "supported (%x:%x)\n", lat_timer,
				      pdev->vendor, pdev->device);
			return -EINVAL;
		}
	}

        /* Configuring PCIe bridge (PEX), if any. */
	if (!bridge || (0x10b5 != bridge->vendor) ||
	    (0x8111 != bridge->device && 0x8112 != bridge->device))
                return 0;

	module_printk (KERN_INFO, "Setting PCIe bridge Cache Line Size to %d\n",
		       cacheline_bridge);
	if (set_cacheline (bridge, cacheline_bridge))
            return -EINVAL;

        return 0;
}

/*
Allocates memory for pmc_e1_controller structure, returns a pointer
to allocated strunture, NULL if allocation insucessful.	*/
static struct pmc_e1_controller * pmc_e1_alloc(void)
{
	struct pmc_e1_controller * tmp_dev;
	//create structure for pmc_device
	tmp_dev = kmalloc(sizeof(* tmp_dev), GFP_KERNEL);
	if(tmp_dev)
	{
		memset(tmp_dev,0, sizeof(*tmp_dev));
	}
	return tmp_dev;
}

static int ptmc_e1_init_one (struct pci_dev *pdev, const struct pci_device_id *ent)
{
	struct pmc_e1_controller * tmp_dev;
	int err;

	if (pci_enable_device(pdev))
		return -EIO;  //last
	if (device_number > (PMC_E18_NR-1))
		return -EBUSY;

	err = tune_pci (pdev);
	if (err)
	    return err;

	tmp_dev = pmc_e1_alloc();
	if (!tmp_dev)
	{
		module_printk(KERN_ERR, "Failed to allocate device.\n");
		return -ENOMEM;
	}
       	tmp_dev->dev_number = device_number;
        sprintf(tmp_dev->name,"quasarm%d", device_number);

	//bind with pci device;
	tmp_dev->pci_device = pdev;
	//add to list of controllers, I don't remember why should I use this list, just part of old code from old project.
	list_add_tail(&tmp_dev->dev_node, &dev_list_head);
      // printk("pciaddr = 0x%lx, irq = %d, ptmc_dev = %p \n", pci_resource_start(pdev,0), pdev->irq, tmp_dev);
	if (request_mem_region (pci_resource_start(pdev,0), WIN_SIZE, tmp_dev->name) == NULL)
	{
	    module_printk (KERN_ERR, "I/O resource 0x%x @ 0x%llx busy\n",
		    WIN_SIZE, (u64) pci_resource_start(pdev,0));
	    goto fail_req_bar0;
	}
	if (request_mem_region (pci_resource_start (pdev, 1), WIN_SIZE, tmp_dev->name) == NULL)
	{
	    module_printk (KERN_ERR, "I/O resource 0x%x @ 0x%llx busy\n",
		    WIN_SIZE, (u64) pci_resource_start(pdev,1));
	    goto fail_req_bar1;
	}
	tmp_dev->bar0 = ioremap_nocache(pci_resource_start(pdev,0), WIN_SIZE);
	if (!tmp_dev->bar0)	{
	    module_printk(KERN_ERR, "Can't map 0x%x @ 0x%llx\n",
		   WIN_SIZE, (u64) pci_resource_start(pdev,0));
	    goto fail_map_bar0;
	}
	tmp_dev->bar1 = ioremap_nocache(pci_resource_start(pdev,1), WIN_SIZE);
	if (!tmp_dev->bar1)	{
	    module_printk(KERN_ERR, "Can't map 0x%x @ 0x%llx\n",
		   WIN_SIZE, (u64) pci_resource_start(pdev,1));
	    goto fail_map_bar1;
	}
	ptmc_sysfs_create (tmp_dev);

	//TODO define how much ports do we have, fill tmp_dev->e1_amount field, for now just think we have two ports.

	tmp_dev->e1_amount = 1;  // 0 - one port, 1 - two ports

	module_printk(KERN_INFO, "Registering interrupt %d with flags 0x%x\n",
		     pdev->irq, irqflags);
	if (request_irq(pdev->irq, ptmc_e1_interrupt_hr,
			irqflags, tmp_dev->name, (void *) tmp_dev) < 0) {
	    module_printk(KERN_ERR, "Can't register interrupt: %d\n", pdev->irq);
	    goto fail_reg_irq;
	}
        cdev_init(&tmp_dev->dev_file, &pmc_e1_fops);
	tmp_dev->dev_file.owner = THIS_MODULE;
	tmp_dev->dev_file.ops = &pmc_e1_fops;
	if (cdev_add(&tmp_dev->dev_file, MKDEV(MAJOR(dev_file_glob),  device_number), 1) )
	{
	    module_printk(KERN_ERR, "Can't register device file for: %d\n", device_number);
	    goto fail_req_file;
	}

	device_number++; //for non-hot plug systems, for hot-plug system we probably should change system of numbering...
	spin_lock_init(&tmp_dev->dev_lock);
	//mark module as initialised
	tmp_dev->registered = 1;

	//Reset and initialise board;

	//initialising board registers
	init_ptmc_regs(tmp_dev);

	//init framer registers
	init_framer_regs(tmp_dev);

	//initialising time-slots maping table.
	init_switch_map(tmp_dev);

	//DAHDI
	module_printk(KERN_INFO, "Quasar-mini board with %d ports found, dev id: %d, hw_fw_rev: 0x%x, (pci slot %d)\n",
		      tmp_dev->e1_amount + 1, tmp_dev->dev_number, tmp_dev->bar0[0x0D], PCI_SLOT(pdev->devfn)+1);
	err = e12zap_register(tmp_dev);
	if(err)
	    module_printk(KERN_ERR, "Quasar-mini dahdi registration error.\n");
        return 0;
	//Clean up all resources, but leave in list.
fail_req_file: free_irq(pdev->irq, (void *) tmp_dev);

fail_reg_irq:
        ptmc_sysfs_delete (tmp_dev);
	iounmap((void *)tmp_dev->bar1);
fail_map_bar1: iounmap((void *)tmp_dev->bar0);
fail_map_bar0: release_mem_region(pci_resource_start(pdev, 1), WIN_SIZE);//unreg 1 bar;
fail_req_bar1: release_mem_region(pci_resource_start(pdev, 0), WIN_SIZE);//unreg 0 bar;
fail_req_bar0: tmp_dev->registered = 0;
return -EBUSY;
}

static void ptmc_e1_remove_one (struct pci_dev *pdev)
{
    struct list_head *pos, *n;
	struct pmc_e1_controller *tmp_dev=NULL;
	list_for_each_safe(pos, n, &dev_list_head)
	{
		if (list_entry(pos,struct pmc_e1_controller,dev_node)->pci_device == pdev)
		{
			tmp_dev = list_entry(pos, struct pmc_e1_controller,dev_node);
		}
	}
	if(tmp_dev)
	{
		if(tmp_dev->registered)
		{
			e12zap_unregister(tmp_dev); //DAHDI
			free_irq(pdev->irq, (void *) tmp_dev);
			cdev_del(&tmp_dev->dev_file);
			ptmc_sysfs_delete (tmp_dev);

			iounmap((void *)tmp_dev->bar0);
			release_mem_region (pci_resource_start(pdev, 0), WIN_SIZE);
			iounmap((void *)tmp_dev->bar1);
			release_mem_region (pci_resource_start(pdev, 1), WIN_SIZE);

		}
		//unlink from device's list
		list_del(&tmp_dev->dev_node);
		kfree(tmp_dev);
	}
}

static struct pci_driver ptmc_e1_driver = {
        name:		DRV_NAME,
	probe:		ptmc_e1_init_one,
	remove:		ptmc_e1_remove_one,
	id_table:	ptmc_e1_tbl,
};

static int __init ptmc_e1_init_module(void)
{
	int result, error;
	module_printk(KERN_INFO, "Quasar-mini interface board driver, Parabel Ltd. http://parabel.ru, "
		      "Version: %s\n", QUASARM_VERSION);

	if (PMC_E1_MAJOR)
	{
		dev_file_glob = MKDEV(PMC_E1_MAJOR,0);
		result = register_chrdev_region(dev_file_glob, PMC_E18_NR, DRV_NAME);
	}
	else
	{
		result = alloc_chrdev_region(&dev_file_glob, 0, PMC_E18_NR, DRV_NAME);
	}
	if (result < 0)
	{
		module_printk(KERN_ERR, "Can't register major number for driver.\n");
		return -EBUSY;
	}
#ifdef PMC_E1_UDEV
       pmc_e1_class = class_create(THIS_MODULE, DRV_NAME);
	if (IS_ERR(pmc_e1_class))
	{
		module_printk(KERN_ERR, "Can't create class " DRV_NAME ".\n");
		error=-PTR_ERR(pmc_e1_class);
		goto fail_reg_class;
	}
#endif


	result = pci_register_driver(&ptmc_e1_driver);
	if(result >= 0)
	{
		return 0;
	}
	else
	{
		module_printk(KERN_ERR, "PCI board registration failed!");
		error=-EBUSY;
	}
	pci_unregister_driver(&ptmc_e1_driver);

#ifdef PMC_E1_UDEV
	class_destroy(pmc_e1_class);
fail_reg_class:
#endif

	unregister_chrdev_region(dev_file_glob, PMC_E18_NR);
	return error;
}

static void __exit ptmc_e1_cleanup_module(void)
{
	pci_unregister_driver (&ptmc_e1_driver);
	unregister_chrdev_region(dev_file_glob, PMC_E18_NR);
	module_printk(KERN_INFO, "Quasar-mini E1 interface board driver unloaded.\n");

#ifdef PMC_E1_UDEV
	class_destroy(pmc_e1_class);
#endif

    return;
}

module_init(ptmc_e1_init_module);
module_exit(ptmc_e1_cleanup_module);

module_param(cacheline, int, 0600);
MODULE_PARM_DESC(cacheline, "PCI Cache Line Size (in units of DWORDs).");
module_param(cacheline_bridge, int, 0600);
MODULE_PARM_DESC(cacheline_bridge, "PCIe bridge Cache Line Size (DWORDs).");
module_param(latency, int, 0400);
MODULE_PARM_DESC(latency, "PCI BUS latency timer.");
module_param(irqflags, int, 0600);
MODULE_PARM_DESC(irqflags, "Interrupt handler flags (passed to request_irq).");

MODULE_AUTHOR("Khelik Mikhail <mixel.net@gmail.com>, Pavel Selivanov <pavel[AT]parabel.ru>");
MODULE_DESCRIPTION("Quasar-mini interface board driver");
MODULE_LICENSE("GPL");
MODULE_VERSION (QUASARM_VERSION);
